from .playback_backend.backend_loader import load_backend
import threading
import time
import os

class AudioPlayer:
    def __init__(self, playback_backend, *,
                 debug_allow_multiple_playbacks: bool = False,
                 sounddevice_blocksize: int = 0) -> None:
        """
        Main AudioPlayer class to create a player object.

        More Information: https://nmstr.github.io/PySoundSphere/

        :param playback_backend: The backend to use for playback ("sounddevice", "pygame", "ffplay")
        :param debug_allow_multiple_playbacks: If multiple playbacks should be allowed at the same time. This is used for tests and should be avoided in production.
        :param sounddevice_blocksize: Sounddevice backend specific option to change the blocksize.
        """
        self._debug_allow_multiple_playbacks = debug_allow_multiple_playbacks
        self._playback_backend = load_backend(playback_backend, sounddevice_blocksize)
        self._callback_function = None
        self._file_path = None
        self._is_paused = False
        self._start_time = 0
        self._pause_time = 0
        self._volume = 0.5
        self._paused_at = 0
        self._paused_seconds = 0

    def load(self, file_path: str) -> None:
        """
        Load the song.
        """
        if not os.path.exists(file_path):
            raise ValueError(f'File "{file_path}" does not exist.')
        self._file_path = file_path

    def play(self) -> None:
        """
        Play the song.
        """
        if self._is_paused:
            self._playback_backend.unpause()
            self._paused_seconds += time.time() - self._paused_at
            self._is_paused = False
        elif self._playback_backend.get_busy() and not self._debug_allow_multiple_playbacks:
            raise RuntimeError('Audio is already playing.')
        else:
            if not self._file_path:
                raise ValueError('No audio file loaded.')
            self._playback_backend.load(self._file_path)
            self._playback_backend.play()
            self._start_time = time.time()
            self._start_monitoring()

    def _start_monitoring(self):
        """
        Start a separate thread to monitor the playback status.
        """
        self._monitor_thread = threading.Thread(target=self._monitor_playback)
        self._monitor_thread.daemon = True
        self._monitor_thread.start()

    def _monitor_playback(self):
        """
        Monitor the playback status and trigger a callback when the song finishes.
        """
        while self._playback_backend.get_busy() or self._is_paused:
            time.sleep(0.1)
        if self._callback_function is not None:
            self._callback_function()

    def pause(self) -> None:
        """
        Pause the song.
        """
        if self._playback_backend.get_busy():
            self._playback_backend.pause()
            self._pause_time = time.time() - self._start_time - self._paused_seconds
            self._paused_at = time.time()
            self._is_paused = True

    def stop(self) -> None:
        """
        Stop the song.
        """
        if self._playback_backend.get_busy():
            self._playback_backend.stop()
        self._is_paused = False
        self._pause_time = 0

    def set_callback_function(self, function = None) -> None:
        """
        Sets the callback function.
        """
        self._callback_function = function

    @property
    def position(self) -> float:
        """
        Position in the song in seconds.
        """
        if self._is_paused:
            return self._pause_time
        elif self._playback_backend.get_busy():
            return time.time() - self._start_time - self._paused_seconds
        else:
            return 0

    @position.setter
    def position(self, position: float) -> None:
        if self._playback_backend.get_busy() or self._is_paused:
            self._playback_backend.stop()
            self._playback_backend.play(start_time=position)
            self._start_time = time.time() - position
            self._paused_seconds = 0
            if self._is_paused:
                self._playback_backend.pause()

    @property
    def volume(self) -> float:
        """
        Volume from 0 to 1.
        """
        return self._volume

    @volume.setter
    def volume(self, volume: float) -> None:
        if not isinstance(volume, float):
            raise TypeError('Volume must be a float.')
        self._playback_backend.set_volume(volume)
        self._volume = volume



# uncomment if you have copied the git repo instead of installing the package and copying the file
#import sys; import os; sys.path.append(os.path.dirname(os.path.dirname(__file__)))
import PySoundSphere
import time

player = PySoundSphere.AudioPlayer("sounddevice")
player.load("path/to/your/audio_file")
player.volume = 0.025
player.play()

time.sleep(10)



# uncomment if you have copied the git repo instead of installing the package and copying the file
#import sys; import os; sys.path.append(os.path.dirname(os.path.dirname(__file__)))
import PySoundSphere

player = PySoundSphere.AudioPlayer("sounddevice")
player.volume = 0.025
state = None

while True:
    print("1) Play song")
    print("2) Pause song")
    print("3) Stop song")
    print("4) Set volume")
    print("5) Set position")
    print("6) Exit")

    try:
        choice = int(input(":"))
    except ValueError:
        print("Invalid input")
        continue

    if choice == 1:
        if state == "Paused":
            player.play()
            state = "Paused"
        elif state == "Playing":
            print("Already playing")
        else:
            print("Enter path to song")
            path = input(":")
            player.load(path)
            player.play()
            state = "Playing"
    elif choice == 2:
        player.pause()
        state = "Paused"
    elif choice == 3:
        player.stop()
        state = None
    elif choice == 4:
        print("Enter volume")
        try:
            vol = float(input(":"))
        except ValueError:
            print("Invalid input")
            continue
        player.volume = vol
    elif choice == 5:
        print("Enter position (in seconds)")
        try:
            pos = int(input(":"))
        except ValueError:
            print("Invalid input")
            continue
        player.position = pos
    elif choice == 6:
        break



def load_backend(backend_name: str, sounddevice_blocksize: int = 0) -> object:
    """
    Load the backend.

    Parameters:
        backend_name (str): Name of the backend. Possible values:
        - sounddevice
        - pygame
        - ffplay
        sounddevice_blocksize (int): The blocksize the sounddevice backend should use

    Returns:
        object: Backend object.
    """
    if backend_name == "sounddevice":
        try:
            from .sounddevice_backend import SounddeviceBackend
            return SounddeviceBackend(blocksize = sounddevice_blocksize)
        except ImportError as e:
            raise ValueError(f'Failed to load backend "{backend_name}" with error: "{e}". Is it installed?')

    elif backend_name == "pygame":
        try:
            from .pygame_backend import PygameBackend
            return PygameBackend()
        except ImportError as e:
            raise ValueError(f'Failed to load backend "{backend_name}" with error: "{e}". Is it installed?')
        
    elif backend_name == "ffplay":
        try:
            from .ffplay_backend import FfplayBackend
            return FfplayBackend()
        except ImportError as e:
            raise ValueError(f'Failed to load backend "{backend_name}" with error: "{e}". Is it installed?')

    else:
        raise ValueError(f'Backend "{backend_name}" not found')



from PySoundSphere.main import AudioPlayer
import pytest

@pytest.fixture(params=['wav', 'mp3', 'flac'])
def audio_file(request):
    format = request.param
    base_file_name = "test/audio_files/75hz_stereo"
    yield base_file_name + "." + format

def test_audio_player_play(audio_file: str) -> None:
    player = AudioPlayer('ffplay', debug_allow_multiple_playbacks=True)
    player.load(audio_file)
    player.play()
    assert player._playback_backend.get_busy() == True
    assert player._is_paused == False
    player.stop()

def test_audio_player_pause(audio_file: str) -> None:
    player = AudioPlayer('ffplay', debug_allow_multiple_playbacks=True)
    player.load(audio_file)
    player.play()
    player.pause()
    assert player._playback_backend.get_busy() == False
    assert player._is_paused == True

def test_audio_player_stop(audio_file: str) -> None:
    player = AudioPlayer('ffplay', debug_allow_multiple_playbacks=True)
    player.load(audio_file)
    player.play()
    player.stop()
    assert player._playback_backend.get_busy() == False
    assert player._is_paused == False

def test_audio_player_position(audio_file: str) -> None:
    player = AudioPlayer('ffplay', debug_allow_multiple_playbacks=True)
    player.load(audio_file)
    player.play()
    assert player.position > 0
    player.stop()

def test_audio_player_volume(audio_file: str) -> None:
    player = AudioPlayer('ffplay', debug_allow_multiple_playbacks=True)
    assert player.volume == 0.5
    player.volume = 0.25
    assert player.volume == 0.25

def test_audio_player_volume_running(audio_file: str) -> None:
    player = AudioPlayer('ffplay', debug_allow_multiple_playbacks=True)
    player.load(audio_file)
    player.play()
    assert player.volume == 0.5
    player.volume = 0.25
    assert player.volume == 0.25
    player.stop()



import sounddevice as sd
import soundfile as sf

class SounddeviceBackend:
    def __init__(self, blocksize: int = 0) -> None:
        self._data = None
        self._original_data = None
        self._samplerate = None
        self._stream = None
        self._volume = 0.5
        self._is_busy = False
        self._block_size = blocksize
    
    def load(self, file_path: str) -> None:
        """
        Load the song.

        Parameters:
            file_path (str): Path to the song.
        """
        self._original_data, self._samplerate = sf.read(file_path, dtype='float32')

    def pause(self) -> None:
        """
        Pause the song.
        """
        if self._stream is not None:
            self._stream.stop()
            self._is_busy = False

    def unpause(self) -> None:
        """
        Unpause the song.
        """
        if self._stream is not None:
            self._stream.start()
            self._is_busy = True

    def play(self, start_time: float = 0) -> None:
        """
        Play the song.

        Parameters:
            start_time (float): Position in the song in seconds.
        """
        start_frame = int(start_time * self._samplerate)
        self._data = self._original_data[start_frame:]
        self._stream = sd.OutputStream(
            callback=self._audio_callback,
            samplerate=self._samplerate,
            channels=self._data.shape[1],
            blocksize=self._block_size)
        self._stream.start()
        self._is_busy = True

    def stop(self) -> None:
        """
        Stop the song.
        """
        if self._stream is not None:
            self._stream.stop()
            self._stream.close()
            self._stream = None
            self._is_busy = False

    def set_volume(self, new_volume: float) -> None:
        """
        Set the volume between 0 and 1.

        Parameters:
            volume (float): Volume between 0 and 1.
        """
        self._volume = new_volume
    
    def get_busy(self) -> bool:
        """
        Check if the song is playing.
        """
        return self._is_busy
    
    def _audio_callback(self, outdata, frames, time, status) -> None:
        """
        Callback function for sounddevice.
        """
        if status:
            print(status)
        if self._data is not None:
            if len(self._data) < frames:
                outdata[:len(self._data)] = self._volume * self._data
                outdata[len(self._data):] = 0
                self._is_busy = False
            else:
                outdata[:] = self._volume * self._data[:frames]
                self._data = self._data[frames:]



import subprocess
import threading
import time

class FfplayBackend:
    def __init__(self) -> None:
        self._file_path = None
        self._playback_process = None
        self._song_started_at = 0
        self._song_stopped_at = 0
        self._song_started_ahead = 0
        self._volume = 5.0
        self._is_busy = False

    def load(self, file_path: str) -> None:
        """
        Load the song.
        """
        self._file_path = file_path

    def play(self, start_time: float = 0) -> None:
        """
        Play the song.

        Parameters:
            start_time (float): Position in the song in seconds.
        """
        self._song_started_at = time.time()
        self._song_started_ahead = start_time
        self._playback_process = subprocess.Popen(['ffplay', '-nodisp', '-autoexit', '-ss', str(start_time), '-volume', str(self._volume), self._file_path],
                                                  stdout=subprocess.DEVNULL,
                                                  stderr=subprocess.DEVNULL)
        self._is_busy = True
        threading.Thread(target=self._playback_done).start()

    def _playback_done(self):
        self._playback_process.wait()
        self._is_busy = False

    def pause(self) -> None:
        """
        Pause the song.
        """
        self._song_stopped_at = time.time()
        self._playback_process.kill()
        self._is_busy = False

    def unpause(self) -> None:
        """
        Unpause the song.
        """
        start_time = self._song_stopped_at - self._song_started_at + self._song_started_ahead
        self.play(start_time)
        self._is_busy = True

    def stop(self) -> None:
        """
        Stop the song.
        """
        self._playback_process.kill()
        self._is_busy = False

    def set_volume(self, volume: float) -> None:
        """
        Set the volume between 0 and 1.

        Parameters:
            volume (float): Volume between 0 and 1.
        """
        volume = max(0, min(round(volume * 100), 100))
        self._volume = volume
        if self._is_busy:
            self.pause()
            self.unpause()

    def get_busy(self) -> bool:
        """
        Check if the song is playing.
        """
        return self._is_busy



import pygame

class PygameBackend:
    def __init__(self) -> None:
        pygame.mixer.init()

    def load(self, file_path: str) -> None:
        """
        Load the song.
        """
        pygame.mixer.music.load(file_path)

    def play(self, start_time: float = 0) -> None:
        """
        Play the song.

        Parameters:
            start_time (float): Position in the song in seconds.
        """
        pygame.mixer.music.play(start=start_time)

    def pause(self) -> None:
        """
        Pause the song.
        """
        pygame.mixer.music.pause()

    def unpause(self) -> None:
        """
        Unpause the song.
        """
        pygame.mixer.music.unpause()

    def stop(self) -> None:
        """
        Stop the song.
        """
        pygame.mixer.music.stop()

    def set_volume(self, volume: float) -> None:
        """
        Set the volume between 0 and 1.

        Parameters:
            volume (float): Volume between 0 and 1.
        """
        pygame.mixer.music.set_volume(volume)

    def get_busy(self) -> bool:
        """
        Check if the song is playing.
        """
        return pygame.mixer.music.get_busy()
